import numpy as np
from tqdm import tqdm

from ttsds.benchmarks.benchmark import Benchmark, BenchmarkCategory, BenchmarkDimension
from ttsds.util.dataset import Dataset


def wada_snr(wav):
    # From: https://gist.github.com/johnmeade/d8d2c67b87cda95cd253f55c21387e75
    # Direct blind estimation of the SNR of a speech signal.
    #
    # Paper on WADA SNR:
    #   http://www.cs.cmu.edu/~robust/Papers/KimSternIS08.pdf
    #
    # This function was adapted from this matlab code:
    #   https://labrosa.ee.columbia.edu/projects/snreval/#9

    # init
    eps = 1e-10
    # next 2 lines define a fancy curve derived from a gamma distribution -- see paper
    db_vals = np.arange(-20, 101)
    g_vals = np.array(
        [
            0.40974774,
            0.40986926,
            0.40998566,
            0.40969089,
            0.40986186,
            0.40999006,
            0.41027138,
            0.41052627,
            0.41101024,
            0.41143264,
            0.41231718,
            0.41337272,
            0.41526426,
            0.4178192,
            0.42077252,
            0.42452799,
            0.42918886,
            0.43510373,
            0.44234195,
            0.45161485,
            0.46221153,
            0.47491647,
            0.48883809,
            0.50509236,
            0.52353709,
            0.54372088,
            0.56532427,
            0.58847532,
            0.61346212,
            0.63954496,
            0.66750818,
            0.69583724,
            0.72454762,
            0.75414799,
            0.78323148,
            0.81240985,
            0.84219775,
            0.87166406,
            0.90030504,
            0.92880418,
            0.95655449,
            0.9835349,
            1.01047155,
            1.0362095,
            1.06136425,
            1.08579312,
            1.1094819,
            1.13277995,
            1.15472826,
            1.17627308,
            1.19703503,
            1.21671694,
            1.23535898,
            1.25364313,
            1.27103891,
            1.28718029,
            1.30302865,
            1.31839527,
            1.33294817,
            1.34700935,
            1.3605727,
            1.37345513,
            1.38577122,
            1.39733504,
            1.40856397,
            1.41959619,
            1.42983624,
            1.43958467,
            1.44902176,
            1.45804831,
            1.46669568,
            1.47486938,
            1.48269965,
            1.49034339,
            1.49748214,
            1.50435106,
            1.51076426,
            1.51698915,
            1.5229097,
            1.528578,
            1.53389835,
            1.5391211,
            1.5439065,
            1.54858517,
            1.55310776,
            1.55744391,
            1.56164927,
            1.56566348,
            1.56938671,
            1.57307767,
            1.57654764,
            1.57980083,
            1.58304129,
            1.58602496,
            1.58880681,
            1.59162477,
            1.5941969,
            1.59693155,
            1.599446,
            1.60185011,
            1.60408668,
            1.60627134,
            1.60826199,
            1.61004547,
            1.61192472,
            1.61369656,
            1.61534074,
            1.61688905,
            1.61838916,
            1.61985374,
            1.62135878,
            1.62268119,
            1.62390423,
            1.62513143,
            1.62632463,
            1.6274027,
            1.62842767,
            1.62945532,
            1.6303307,
            1.63128026,
            1.63204102,
        ]
    )

    # peak normalize, get magnitude, clip lower bound
    wav = np.array(wav)
    wav = wav / abs(wav).max()
    abs_wav = abs(wav)
    abs_wav[abs_wav < eps] = eps

    # calcuate statistics
    # E[|z|]
    v1 = max(eps, abs_wav.mean())
    # E[log|z|]
    v2 = np.log(abs_wav).mean()
    # log(E[|z|]) - E[log(|z|)]
    v3 = np.log(v1) - v2

    # table interpolation
    wav_snr_idx = None
    if any(g_vals < v3):
        wav_snr_idx = np.where(g_vals < v3)[0].max()
    # handle edge cases or interpolate
    if wav_snr_idx is None:
        wav_snr = db_vals[0]
    elif wav_snr_idx == len(db_vals) - 1:
        wav_snr = db_vals[-1]
    else:
        wav_snr = db_vals[wav_snr_idx] + (v3 - g_vals[wav_snr_idx]) / (
            g_vals[wav_snr_idx + 1] - g_vals[wav_snr_idx]
        ) * (db_vals[wav_snr_idx + 1] - db_vals[wav_snr_idx])

    # Calculate SNR
    dEng = sum(wav ** 2)
    dFactor = 10 ** (wav_snr / 10)
    dNoiseEng = dEng / (1 + dFactor)  # Noise energy
    dSigEng = dEng * dFactor / (1 + dFactor)  # Signal energy
    snr = 10 * np.log10(dSigEng / dNoiseEng)

    return snr


class WadaSNRBenchmark(Benchmark):
    """
    Benchmark class for the Wada SNR benchmark.
    """

    def __init__(
        self,
        window_size: int = 3,
        step_size: int = 1,
    ):
        super().__init__(
            name="Wada SNR",
            category=BenchmarkCategory.ENVIRONMENT,
            dimension=BenchmarkDimension.ONE_DIMENSIONAL,
            description="The Signal-to-Noise Ratio (SNR) of Wada SNR.",
            window_size=window_size,
            step_size=step_size,
        )
        self.window_size = window_size
        self.step_size = step_size

    def _get_distribution(self, dataset: Dataset) -> np.ndarray:
        """
        Compute the Signal-to-Noise Ratio (SNR) distribution of the Wada SNR model.

        Args:
            dataset (Dataset): The dataset to compute the SNR on.

        Returns:
            float: The Signal-to-Noise Ratio (SNR) distribution of the Wada SNR model.
        """
        snrs = []
        for wav, _ in tqdm(dataset, desc=f"computing snr for {self.name}"):
            # calculate for each 3 seconds, with a 1 second step
            win = self.window_size * dataset.sample_rate
            step = self.step_size * dataset.sample_rate
            while len(wav) > win:
                snr = wada_snr(wav[:win])
                if np.isnan(snr):
                    snr = 0
                snrs.append(snr)
                wav = wav[step:]
                if len(wav) == 0:
                    break
        return np.array(snrs)



"""
This file contains the Benchmark abstract class.
"""

from abc import ABC, abstractmethod
from enum import Enum
import hashlib
import importlib.resources
import json
from typing import List, Union
from functools import lru_cache
from pathlib import Path

import numpy as np

from ttsds.util.dataset import Dataset, DataDistribution
from ttsds.util.cache import cache, load_cache, check_cache, hash_md5
from ttsds.util.distances import wasserstein_distance, frechet_distance


class BenchmarkCategory(Enum):
    """
    Enum class for the different categories of benchmarks.
    """

    OVERALL = 1
    PROSODY = 2
    ENVIRONMENT = 3
    SPEAKER = 4
    PHONETICS = 5
    INTELLIGIBILITY = 6
    TRAINABILITY = 7
    EXTERNAL = 8


class BenchmarkDimension(Enum):
    """
    Enum class for the different dimensions of benchmarks.
    """

    ONE_DIMENSIONAL = 1
    N_DIMENSIONAL = 2


class Benchmark(ABC):
    """
    Abstract class for a benchmark.
    """

    def __init__(
        self,
        name: str,
        category: BenchmarkCategory,
        dimension: BenchmarkDimension,
        description: str,
        **kwargs,
    ):
        self.name = name
        self.key = name.lower().replace(" ", "_")
        self.category = category
        self.dimension = dimension
        self.description = description
        self.kwargs = kwargs

    def get_distribution(self, dataset: Union[Dataset, DataDistribution]) -> np.ndarray:
        """
        Abstract method to get the distribution of the benchmark.
        If the benchmark is one-dimensional, the method should return a
        numpy array with the values of the benchmark for each sample in the dataset.
        If the benchmark is n-dimensional, the method should return a numpy array
        with the values of the benchmark for each sample in the dataset, where each
        row corresponds to a sample and each column corresponds to a dimension of the benchmark.
        """
        ds_hash = hash_md5(dataset)
        benchmark_hash = hash_md5(self)
        cache_name = f"benchmarks/{self.name}/{ds_hash}_{benchmark_hash}"
        if check_cache(cache_name):
            return load_cache(cache_name)
        if check_cache(cache_name + "_mu") and check_cache(cache_name + "_sig"):
            mu = load_cache(cache_name + "_mu")
            sig = load_cache(cache_name + "_sig")
            return (mu, sig)
        if isinstance(dataset, DataDistribution) and self.dimension == BenchmarkDimension.N_DIMENSIONAL:
            mu, sig = dataset.get_distribution(self.key)
            cache(mu, cache_name + "_mu")
            cache(sig, cache_name + "_sig")
            return (mu, sig)
        elif isinstance(dataset, DataDistribution) and self.dimension == BenchmarkDimension.ONE_DIMENSIONAL:
            distribution = dataset.get_distribution(self.key)
            cache(distribution, cache_name)
            return distribution
        distribution = self._get_distribution(dataset)
        cache(distribution, cache_name)
        return distribution

    @abstractmethod
    def _get_distribution(self, dataset: Dataset) -> np.ndarray:
        """
        Abstract method to get the distribution of the benchmark.
        """
        raise NotImplementedError

    def __str__(self) -> str:
        return f"{self.category.name}/{self.name}"

    def __repr__(self):
        return f"{self.category.name}/{self.name}"

    def __hash__(self) -> int:
        h = hashlib.md5()
        h.update(self.name.encode())
        h.update(self.category.name.encode())
        h.update(self.dimension.name.encode())
        h.update(self.description.encode())
        # convert the kwargs to strings
        kwargs_str = {
            k: str(v) if not isinstance(v, dict) else json.dumps(v, sort_keys=True)
            for k, v in self.kwargs.items()
        }
        h.update(json.dumps(kwargs_str, sort_keys=True).encode())
        return int(h.hexdigest(), 16)

    @lru_cache(maxsize=None)
    def compute_distance(
        self,
        one_dataset: Union[Dataset, DataDistribution],
        other_dataset: Union[Dataset, DataDistribution],
    ) -> float:
        """
        Compute the distance between the distributions of the benchmark in two datasets.
        """
        one_distribution = self.get_distribution(one_dataset)
        other_distribution = self.get_distribution(other_dataset)
        if self.dimension == BenchmarkDimension.ONE_DIMENSIONAL:
            return wasserstein_distance(one_distribution, other_distribution)
        elif self.dimension == BenchmarkDimension.N_DIMENSIONAL:
            return frechet_distance(one_distribution, other_distribution)
        else:
            raise ValueError("Invalid benchmark dimension")

    def compute_score(
        self,
        dataset: Dataset,
        reference_datasets: List[Dataset],
        noise_datasets: List[Dataset],
    ) -> float:
        """
        Compute the score of the benchmark on a dataset.
        """
        noise_scores = []
        for noise_ds in noise_datasets:
            score = self.compute_distance(noise_ds, dataset)
            noise_scores.append(score)
        noise_scores = np.array(noise_scores)

        dataset_scores = []
        for ref_ds in reference_datasets:
            score = self.compute_distance(ref_ds, dataset)
            dataset_scores.append(score)
        dataset_scores = np.array(dataset_scores)

        closest_noise_idx = np.argmin(noise_scores)
        closest_dataset_idx = np.argmin(dataset_scores)

        print(f"Closest noise dataset: {noise_datasets[closest_noise_idx].name}")
        print(
            f"Closest reference dataset: {reference_datasets[closest_dataset_idx].name}"
        )

        noise_score = np.min(noise_scores)
        dataset_score = np.min(dataset_scores)
        combined_score = dataset_score + noise_score
        score = (noise_score / combined_score) * 100
        # TODO: compute confidence interval
        return (
            score,
            1.0,
            (
                noise_datasets[closest_noise_idx].name,
                reference_datasets[closest_dataset_idx].name,
            ),
        )



import torch
import torch.nn as nn
import fairseq
import os
import hydra


def load_ssl_model(cp_path):
    ssl_model_type = cp_path.split("/")[-1]
    wavlm = "WavLM" in ssl_model_type
    if wavlm:
        checkpoint = torch.load(cp_path)
        cfg = WavLMConfig(checkpoint["cfg"])
        ssl_model = WavLM(cfg)
        ssl_model.load_state_dict(checkpoint["model"])
        if "Large" in ssl_model_type:
            SSL_OUT_DIM = 1024
        else:
            SSL_OUT_DIM = 768
    else:
        if ssl_model_type == "wav2vec_small.pt":
            SSL_OUT_DIM = 768
        elif ssl_model_type in ["w2v_large_lv_fsh_swbd_cv.pt", "xlsr_53_56k.pt"]:
            SSL_OUT_DIM = 1024
        else:
            print("*** ERROR *** SSL model type " + ssl_model_type + " not supported.")
            exit()
        model, cfg, task = fairseq.checkpoint_utils.load_model_ensemble_and_task(
            [cp_path]
        )
        ssl_model = model[0]
        ssl_model.remove_pretraining_modules()
    return SSL_model(ssl_model, SSL_OUT_DIM, wavlm)


class SSL_model(nn.Module):
    def __init__(self, ssl_model, ssl_out_dim, wavlm) -> None:
        super(SSL_model, self).__init__()
        self.ssl_model, self.ssl_out_dim = ssl_model, ssl_out_dim
        self.WavLM = wavlm

    def forward(self, batch):
        wav = batch["wav"]
        wav = wav.squeeze(1)  # [batches, audio_len]
        if self.WavLM:
            x = self.ssl_model.extract_features(wav)[0]
        else:
            res = self.ssl_model(wav, mask=False, features_only=True)
            x = res["x"]
        return {"ssl-feature": x}

    def get_output_dim(self):
        return self.ssl_out_dim


class PhonemeEncoder(nn.Module):
    """
    PhonemeEncoder consists of an embedding layer, an LSTM layer, and a linear layer.
    Args:
        vocab_size: the size of the vocabulary
        hidden_dim: the size of the hidden state of the LSTM
        emb_dim: the size of the embedding layer
        out_dim: the size of the output of the linear layer
        n_lstm_layers: the number of LSTM layers
    """

    def __init__(
        self,
        vocab_size,
        hidden_dim,
        emb_dim,
        out_dim,
        n_lstm_layers,
        with_reference=True,
    ) -> None:
        super().__init__()
        self.with_reference = with_reference
        self.embedding = nn.Embedding(vocab_size, emb_dim)
        self.encoder = nn.LSTM(
            emb_dim,
            hidden_dim,
            num_layers=n_lstm_layers,
            dropout=0.1,
            bidirectional=True,
        )
        self.linear = nn.Sequential(
            nn.Linear(hidden_dim + hidden_dim * self.with_reference, out_dim), nn.ReLU()
        )
        self.out_dim = out_dim

    def forward(self, batch):
        seq = batch["phonemes"]
        lens = batch["phoneme_lens"]
        reference_seq = batch["reference"]
        reference_lens = batch["reference_lens"]
        emb = self.embedding(seq)
        emb = torch.nn.utils.rnn.pack_padded_sequence(
            emb, lens, batch_first=True, enforce_sorted=False
        )
        _, (ht, _) = self.encoder(emb)
        feature = ht[-1] + ht[0]
        if self.with_reference:
            if reference_seq == None or reference_lens == None:
                raise ValueError(
                    "reference_batch and reference_lens should not be None when with_reference is True"
                )
            reference_emb = self.embedding(reference_seq)
            reference_emb = torch.nn.utils.rnn.pack_padded_sequence(
                reference_emb, reference_lens, batch_first=True, enforce_sorted=False
            )
            _, (ht_ref, _) = self.encoder(emb)
            reference_feature = ht_ref[-1] + ht_ref[0]
            feature = self.linear(torch.cat([feature, reference_feature], 1))
        else:
            feature = self.linear(feature)
        return {"phoneme-feature": feature}

    def get_output_dim(self):
        return self.out_dim


class DomainEmbedding(nn.Module):
    def __init__(self, n_domains, domain_dim) -> None:
        super().__init__()
        self.embedding = nn.Embedding(n_domains, domain_dim)
        self.output_dim = domain_dim

    def forward(self, batch):
        return {"domain-feature": self.embedding(batch["domains"])}

    def get_output_dim(self):
        return self.output_dim


class LDConditioner(nn.Module):
    """
    Conditions ssl output by listener embedding
    """

    def __init__(self, input_dim, judge_dim, num_judges=None):
        super().__init__()
        self.input_dim = input_dim
        self.judge_dim = judge_dim
        self.num_judges = num_judges
        assert num_judges != None
        self.judge_embedding = nn.Embedding(num_judges, self.judge_dim)
        # concat [self.output_layer, phoneme features]

        self.decoder_rnn = nn.LSTM(
            input_size=self.input_dim + self.judge_dim,
            hidden_size=512,
            num_layers=1,
            batch_first=True,
            bidirectional=True,
        )  # linear?
        self.out_dim = self.decoder_rnn.hidden_size * 2

    def get_output_dim(self):
        return self.out_dim

    def forward(self, x, batch):
        judge_ids = batch["judge_id"]
        if "phoneme-feature" in x.keys():
            concatenated_feature = torch.cat(
                (
                    x["ssl-feature"],
                    x["phoneme-feature"]
                    .unsqueeze(1)
                    .expand(-1, x["ssl-feature"].size(1), -1),
                ),
                dim=2,
            )
        else:
            concatenated_feature = x["ssl-feature"]
        if "domain-feature" in x.keys():
            concatenated_feature = torch.cat(
                (
                    concatenated_feature,
                    x["domain-feature"]
                    .unsqueeze(1)
                    .expand(-1, concatenated_feature.size(1), -1),
                ),
                dim=2,
            )
        if judge_ids != None:
            concatenated_feature = torch.cat(
                (
                    concatenated_feature,
                    self.judge_embedding(judge_ids)
                    .unsqueeze(1)
                    .expand(-1, concatenated_feature.size(1), -1),
                ),
                dim=2,
            )
            decoder_output, (h, c) = self.decoder_rnn(concatenated_feature)
        return decoder_output


class Projection(nn.Module):
    def __init__(self, input_dim, hidden_dim, activation, range_clipping=False):
        super(Projection, self).__init__()
        self.range_clipping = range_clipping
        output_dim = 1
        if range_clipping:
            self.proj = nn.Tanh()

        self.net = nn.Sequential(
            nn.Linear(input_dim, hidden_dim),
            activation,
            nn.Dropout(0.3),
            nn.Linear(hidden_dim, output_dim),
        )
        self.output_dim = output_dim

    def forward(self, x, batch):
        output = self.net(x)

        # range clipping
        if self.range_clipping:
            return self.proj(output) * 2.0 + 3
        else:
            return output

    def get_output_dim(self):
        return self.output_dim



from tqdm import tqdm
from wvmos import get_wvmos
import tempfile
import soundfile as sf
import numpy as np

from ttsds.benchmarks.benchmark import Benchmark, BenchmarkCategory, BenchmarkDimension
from ttsds.util.dataset import Dataset


class WVMOSBenchmark(Benchmark):
    """
    Benchmark class for the Hubert benchmark.
    """

    def __init__(
        self,
    ):
        super().__init__(
            name="WVMOS",
            category=BenchmarkCategory.EXTERNAL,
            dimension=BenchmarkDimension.ONE_DIMENSIONAL,
            description="The WVMOS benchmark.",
        )
        self.model = get_wvmos()

    def _get_distribution(self, dataset: Dataset) -> np.ndarray:
        """
        Get the distribution of the WVMOS benchmark.

        Args:
            dataset (DirectoryDataset): The dataset to get the distribution from.

        Returns:
            np.ndarray: The distribution of the WVMOS benchmark.
        """
        scores = []
        for wav, txt in tqdm(dataset, desc=f"computing scores for {self.name}"):
            with tempfile.NamedTemporaryFile(suffix=".wav") as f:
                sf.write(f.name, wav, dataset.sample_rate)
                score = self.model.calculate_one(f.name)
            scores.append(score)
        return np.array(scores)



from typing import List, Optional
import importlib.resources
from time import time
from pathlib import Path
import pickle
import gzip
import requests

import pandas as pd
from transformers import logging
import numpy as np
from sklearn.decomposition import PCA

from ttsds.benchmarks.environment.voicefixer import VoiceFixerBenchmark
from ttsds.benchmarks.environment.wada_snr import WadaSNRBenchmark
from ttsds.benchmarks.general.hubert import HubertBenchmark
from ttsds.benchmarks.general.wav2vec2 import Wav2Vec2Benchmark
from ttsds.benchmarks.general.wavlm import WavLMBenchmark
from ttsds.benchmarks.intelligibility.w2v2_wer import Wav2Vec2WERBenchmark
from ttsds.benchmarks.intelligibility.whisper_wer import WhisperWERBenchmark
from ttsds.benchmarks.prosody.mpm import MPMBenchmark
from ttsds.benchmarks.prosody.pitch import PitchBenchmark
from ttsds.benchmarks.prosody.hubert_token import HubertTokenBenchmark
from ttsds.benchmarks.speaker.wespeaker import WeSpeakerBenchmark
from ttsds.benchmarks.speaker.dvector import DVectorBenchmark
from ttsds.benchmarks.benchmark import BenchmarkCategory, BenchmarkDimension
from ttsds.util.dataset import Dataset, TarDataset, DataDistribution, DEFAULT_BENCHMARKS

# we do this to avoid "some weights of the model checkpoint at ... were not used when initializing" warnings
logging.set_verbosity_error()


benchmark_dict = {
    "hubert": HubertBenchmark,
    "wav2vec2": Wav2Vec2Benchmark,
    "wavlm": WavLMBenchmark,
    "wav2vec2_wer": Wav2Vec2WERBenchmark,
    "whisper_wer": WhisperWERBenchmark,
    "mpm": MPMBenchmark,
    "pitch": PitchBenchmark,
    "wespeaker": WeSpeakerBenchmark,
    "dvector": DVectorBenchmark,
    "hubert_token": HubertTokenBenchmark,
    "voicefixer": VoiceFixerBenchmark,
    "wada_snr": WadaSNRBenchmark,
}

with importlib.resources.path("ttsds", "data") as data_path:
    # if they don't exist, download from github
    for noise_name in [
        "esc50",
        "all_ones",
        "all_zeros",
        "normal_distribution",
        "uniform_distribution",
    ]:
        if not Path(f"{data_path}/noise_{noise_name}.pkl.gz").exists():
            print(f"Downloading noise_{noise_name}.pkl.gz")
            url = f"https://github.com/ttsds/ttsds/raw/main/src/ttsds/data/noise_{noise_name}.pkl.gz"
            r = requests.get(url)
            with open(f"{data_path}/noise_{noise_name}.pkl.gz", "wb") as f:
                f.write(r.content)

    for speech_name in [
        "blizzard2008",
        "blizzard2013",
        "common_voice",
        "libritts_test",
        "libritts_r_test",
        "lj_speech",
        "vctk",
    ]:
        if not Path(f"{data_path}/reference_speech_{speech_name}.pkl.gz").exists():
            print(f"Downloading reference_speech_{speech_name}.pkl.gz")
            url = f"https://github.com/ttsds/ttsds/raw/main/src/ttsds/data/reference_speech_{speech_name}.pkl.gz"
            r = requests.get(url)
            with open(f"{data_path}/reference_speech_{speech_name}.pkl.gz", "wb") as f:
                f.write(r.content)


    # check if the reference and noise distributions are already saved
    if not Path(f"{data_path}/reference_speech_blizzard2008.pkl.gz").exists():
        print("Creating reference distributions")
        ref_benchmark_dict = {
            k: v() for k, v in benchmark_dict.items()
        }
        REFERENCE_DISTS = [
            DataDistribution(
                TarDataset(data_path / "original" / f"speech_{name}.tar.gz"),
                ref_benchmark_dict,
                benchmarks=DEFAULT_BENCHMARKS,
                name=f"speech_{name}",
            )
            for name in [
                "blizzard2008",
                "blizzard2013",
                "common_voice",
                "libritts_test",
                "libritts_r_test",
                "lj_speech",
                "vctk",
            ]
        ]
        # save the reference distributions
        for dist in REFERENCE_DISTS:
            dist.to_pickle(f"{data_path}/reference_{dist.name}.pkl.gz")

    REFERENCE_DISTS = [
        DataDistribution.from_pickle(f"{data_path}/reference_{name}.pkl.gz")
        for name in [
            "speech_blizzard2008",
            "speech_blizzard2013",
            "speech_common_voice",
            "speech_libritts_test",
            "speech_libritts_r_test",
            "speech_lj_speech",
            "speech_vctk",
        ]
    ]

    if not Path(f"{data_path}/noise_esc50.pkl.gz").exists():
        print("Creating noise distributions")
        ref_benchmark_dict = {
            k: v() for k, v in benchmark_dict.items()
        }
        NOISE_DISTS = [
            DataDistribution(
                TarDataset(data_path / "original" / f"noise_{name}.tar.gz"),
                ref_benchmark_dict,
                benchmarks=DEFAULT_BENCHMARKS,
                name=name,
            )
            for name in [
                "esc50",
                "all_ones",
                "all_zeros",
                "normal_distribution",
                "uniform_distribution",
            ]
        ]
        # save the noise distributions
        for dist in NOISE_DISTS:
            dist.to_pickle(f"{data_path}/noise_{dist.name}.pkl.gz")

    NOISE_DISTS = [
        DataDistribution.from_pickle(f"{data_path}/noise_{name}.pkl.gz")
        for name in [
            "esc50",
            "all_ones",
            "all_zeros",
            "normal_distribution",
            "uniform_distribution",
        ]
    ]


class BenchmarkSuite:
    def __init__(
        self,
        datasets: List[Dataset],
        benchmarks: List[str] = DEFAULT_BENCHMARKS,
        print_results: bool = True,
        skip_errors: bool = False,
        noise_distributions: List[DataDistribution] = NOISE_DISTS,
        reference_distributions: List[DataDistribution] = REFERENCE_DISTS,
        write_to_file: str = None,
    ):
        self.benchmarks = benchmarks
        self.benchmark_objects = [benchmark_dict[benchmark]() for benchmark in benchmarks]
        # sort by category and then by name
        self.benchmark_objects = sorted(
            self.benchmark_objects, key=lambda x: (x.category.value, x.name)
        )
        self.datasets = datasets
        self.datasets = sorted(self.datasets, key=lambda x: x.name)
        self.database = pd.DataFrame(
            columns=[
                "benchmark_name",
                "benchmark_category",
                "dataset",
                "score",
                "ci",
                "time_taken",
                "noise_dataset",
                "reference_dataset",
            ]
        )
        self.print_results = print_results
        self.skip_errors = skip_errors
        self.noise_distributions = noise_distributions
        self.reference_distributions = reference_distributions
        self.write_to_file = write_to_file
        if Path(write_to_file).exists():
            self.database = pd.read_csv(write_to_file, index_col=0)
            self.database = self.database.reset_index()

    def run(self) -> pd.DataFrame:
        for benchmark in self.benchmark_objects:
            for dataset in self.datasets:
                # empty lines for better readability
                print("\n")
                print(f"{'='*80}")
                print(f"Benchmark Category: {benchmark.category.name}")
                print(f"Running {benchmark.name} on {dataset.root_dir}")
                try:
                    # check if it's in the database
                    if (
                        (self.database["benchmark_name"] == benchmark.name)
                        & (self.database["dataset"] == dataset.name)
                    ).any():
                        print(
                            f"Skipping {benchmark.name} on {dataset.name} as it's already in the database"
                        )
                        continue
                    start = time()
                    if "WER".lower() in benchmark.name.lower():
                        print([
                            x.get_distribution(benchmark.key) for x in self.reference_distributions
                        ])
                        print([
                            x.get_distribution(benchmark.key) for x in self.noise_distributions
                        ])
                        print(benchmark.get_distribution(dataset))
                    score = benchmark.compute_score(
                        dataset, self.reference_distributions, self.noise_distributions
                    )
                    time_taken = time() - start
                except Exception as e:
                    if self.skip_errors:
                        print(f"Error: {e}")
                        score = (np.nan, np.nan)
                        time_taken = np.nan
                    else:
                        raise e
                result = {
                    "benchmark_name": [benchmark.name],
                    "benchmark_category": [benchmark.category.value],
                    "dataset": [dataset.name],
                    "score": [score[0]],
                    "ci": [score[1]],
                    "time_taken": [time_taken],
                    "noise_dataset": [score[2][0]],
                    "reference_dataset": [score[2][1]],
                }
                if self.print_results:
                    print(result)
                self.database = pd.concat(
                    [
                        self.database,
                        pd.DataFrame(result),
                    ],
                    ignore_index=True,
                )
                if self.write_to_file is not None:
                    self.database["score"] = self.database["score"].astype(float)
                    self.database = self.database.sort_values(
                        ["benchmark_category", "benchmark_name", "score"],
                        ascending=[True, True, False],
                    )
                    self.database.to_csv(self.write_to_file, index=False)
        return self.database

    @staticmethod
    def aggregate_df(df: pd.DataFrame) -> pd.DataFrame:
        def concat_text(x):
            return ", ".join(x)

        df["benchmark_category"] = df["benchmark_category"].apply(
            lambda x: BenchmarkCategory(x).name
        )
        df = (
            df.groupby(
                [
                    "benchmark_category",
                    "dataset",
                ]
            )
            .agg(
                {
                    "score": ["mean"],
                    "ci": ["mean"],
                    "time_taken": ["mean"],
                    "noise_dataset": [concat_text],
                    "reference_dataset": [concat_text],
                    "benchmark_name": [concat_text],
                }
            )
            .reset_index()
        )
        # remove multiindex
        df.columns = [x[0] for x in df.columns.ravel()]
        # drop the benchmark_name column
        df = df.drop("benchmark_name", axis=1)
        # replace benchmark_category number with string
        return df

    def get_aggregated_results(self) -> pd.DataFrame:
        df = self.database.copy()
        return BenchmarkSuite.aggregate_df(df)

    def get_benchmark_distribution(
        self,
        benchmark_name: str,
        dataset_name: str,
        pca_components: Optional[int] = None,
    ) -> dict:
        benchmark = [x for x in self.benchmark_objects if x.name == benchmark_name][0]
        dataset = [x for x in self.datasets if x.name == dataset_name][0]
        closest_noise = self.database[
            (self.database["benchmark_name"] == benchmark_name)
            & (self.database["dataset"] == dataset_name)
        ]["noise_dataset"].values[0]
        closest_noise = [
            x for x in self.noise_distributions if x.name == closest_noise
        ][0]
        other_noise = [
            x for x in self.noise_distributions if x.name != closest_noise.name
        ][0]
        closest_reference = self.database[
            (self.database["benchmark_name"] == benchmark_name)
            & (self.database["dataset"] == dataset_name)
        ]["reference_dataset"].values[0]
        closest_reference = [
            x for x in self.reference_distributions if x.name == closest_reference
        ][0]
        other_reference = [
            x for x in self.reference_distributions if x.name != closest_reference.name
        ][0]
        result = {
            "benchmark_distribution": benchmark.get_distribution(dataset),
            "noise_distribution": benchmark.get_distribution(closest_noise),
            "reference_distribution": benchmark.get_distribution(closest_reference),
            "other_noise_distribution": benchmark.get_distribution(other_noise),
            "other_reference_distribution": benchmark.get_distribution(other_reference),
        }
        if pca_components is not None:
            pca = PCA(n_components=pca_components)
            # fit on all except the benchmark distribution
            pca.fit(
                np.vstack(
                    [v for k, v in result.items() if k != "benchmark_distribution"]
                )
            )
            result = {k: pca.transform(v) for k, v in result.items()}
        return result



import os
from pathlib import Path

import requests
import numpy as np
import hydra
import pytorch_lightning as pl
import torch
import torch.nn as nn

from ttsds.benchmarks.external.utmos.model import (
    load_ssl_model,
    PhonemeEncoder,
    DomainEmbedding,
    LDConditioner,
    Projection,
)
from ttsds.util.cache import CACHE_DIR


class BaselineLightningModule(pl.LightningModule):
    def __init__(self, cfg):
        super().__init__()
        self.cfg = cfg
        self.construct_model()
        self.save_hyperparameters()

    def construct_model(self):
        wav2vec_path = Path(CACHE_DIR) / "wav2vec_small.pt"
        url = "https://huggingface.co/spaces/sarulab-speech/UTMOS-demo/resolve/main/wav2vec_small.pt?download=true"
        if not wav2vec_path.exists():
            print("Downloading wav2vec_small.pt...")
            response = requests.get(url)
            with open(wav2vec_path, "wb") as f:
                f.write(response.content)
        wav2vec_path = str(wav2vec_path)
        self.feature_extractors = nn.ModuleList(
            [
                load_ssl_model(cp_path=wav2vec_path),
                DomainEmbedding(3, 128),
            ]
        )
        output_dim = sum(
            [
                feature_extractor.get_output_dim()
                for feature_extractor in self.feature_extractors
            ]
        )
        output_layers = [
            LDConditioner(judge_dim=128, num_judges=3000, input_dim=output_dim)
        ]
        output_dim = output_layers[-1].get_output_dim()
        output_layers.append(
            Projection(
                hidden_dim=2048,
                activation=torch.nn.ReLU(),
                range_clipping=False,
                input_dim=output_dim,
            )
        )

        self.output_layers = nn.ModuleList(output_layers)

    def forward(self, inputs):
        outputs = {}
        for feature_extractor in self.feature_extractors:
            outputs.update(feature_extractor(inputs))
        x = outputs
        for output_layer in self.output_layers:
            x = output_layer(x, inputs)
        return x



import re

from transformers import Wav2Vec2Processor, Wav2Vec2ForCTC
from tqdm import tqdm
from jiwer import wer
import torch
import numpy as np
import librosa

from ttsds.benchmarks.benchmark import Benchmark, BenchmarkCategory, BenchmarkDimension
from ttsds.util.dataset import Dataset


class Wav2Vec2WERBenchmark(Benchmark):
    """
    Benchmark class for the Wav2Vec2 Word Error Rate (WER) benchmark.
    """

    def __init__(
        self,
        wav2vec2_model: str = "facebook/wav2vec2-base-960h",
    ):
        super().__init__(
            name="Wav2Vec2 WER",
            category=BenchmarkCategory.INTELLIGIBILITY,
            dimension=BenchmarkDimension.ONE_DIMENSIONAL,
            description="The Word Error Rate (WER) of Wav2Vec2.",
            wav2vec2_model=wav2vec2_model,
        )
        self.processor = Wav2Vec2Processor.from_pretrained(wav2vec2_model)
        self.model = Wav2Vec2ForCTC.from_pretrained(wav2vec2_model)
        self.model.eval()

    def _get_distribution(self, dataset: Dataset) -> np.ndarray:
        """
        Compute the Word Error Rate (WER) distribution of the Wav2Vec2 model.

        Args:
            dataset (Dataset): The dataset to compute the WER on.

        Returns:
            float: The Word Error Rate (WER) distribution of the Wav2Vec2 model.
        """
        wers = []
        for wav, gt_transcript in tqdm(dataset, desc=f"computing WER for {self.name}"):
            if dataset.sample_rate != 16000:
                wav = librosa.resample(
                    wav, orig_sr=dataset.sample_rate, target_sr=16000
                )
            input_values = self.processor(
                wav, return_tensors="pt", sampling_rate=16000
            ).input_values
            with torch.no_grad():
                logits = self.model(input_values).logits
            predicted_ids = torch.argmax(logits, dim=-1)
            pred_transcript = self.processor.batch_decode(predicted_ids)[0]
            pred_transcript = re.sub(r"[^\w\s]", "", pred_transcript)
            gt_transcript = re.sub(r"[^\w\s]", "", gt_transcript)
            pred_transcript = re.sub(r"\s+", " ", pred_transcript).lower()
            gt_transcript = re.sub(r"\s+", " ", gt_transcript).lower()
            wers.append(wer(gt_transcript, pred_transcript))
        return np.array(wers)



"""
The `DirectoryDataset` class is a dataset class for a directory containing wav files and corresponding text files.
"""

from abc import ABC, abstractmethod
from copy import deepcopy
import hashlib
from pathlib import Path
import tarfile
from typing import Tuple, List, Dict
import pickle
import gzip

import numpy as np
import librosa

from ttsds.util.cache import cache, check_cache, load_cache, hash_md5


class Dataset(ABC):
    """
    Abstract class for a dataset.
    """

    def __init__(self, name, sample_rate: int = 22050):
        self.sample_rate = sample_rate
        self.wavs = []
        self.texts = []
        self.sample_params = {
            "n": None,
            "seed": None,
        }
        self.name = name
        self.indices = None

    @abstractmethod
    def __len__(self) -> int:
        """
        Get the number of samples in the dataset.

        Returns:
            int: The number of samples in the dataset.
        """
        raise NotImplementedError

    @abstractmethod
    def __getitem__(self, idx: int) -> Tuple[np.ndarray, str]:
        """
        Get a sample from the dataset.

        Args:
            idx (int): The index of the sample.

        Returns:
            Tuple[Path, Path, str]: A tuple containing the wav file and text file.
        """
        raise NotImplementedError

    def sample(self, n: int, seed: int = 42) -> "DirectoryDataset":
        """
        Sample n samples from the dataset.

        Args:
            n (int): The number of samples to sample.
            seed (int): The seed for the random number generator.

        Returns:
            DirectoryDataset: A sampled dataset.
        """
        rng = np.random.default_rng(seed)
        self.indices = rng.choice(len(self), size=n, replace=False)
        self.sample_params = {"n": n, "seed": seed}
        return self


class DirectoryDataset(Dataset):
    """
    A dataset class for a directory containing
    with wav files and corresponding text files.
    """

    def __init__(self, root_dir: str = None, sample_rate: int = 22050):
        super().__init__(Path(root_dir).name, sample_rate)
        if root_dir is None:
            raise ValueError("root_dir must be provided.")
        self.root_dir = Path(root_dir)
        # we assume that the root directory contains
        wavs, texts = [], []
        for wav_file in Path(root_dir).rglob("*.wav"):
            wavs.append(wav_file)
            text = wav_file.with_suffix(".txt")
            texts.append(text)
        self.wavs = wavs
        self.texts = texts

    def __len__(self) -> int:
        if self.indices is not None:
            return len(self.indices)
        return len(self.wavs)

    def __getitem__(self, idx: int) -> Tuple[np.ndarray, str]:
        if self.indices is not None:
            idx = self.indices[idx]
        wav, sr = self.wavs[idx], self.sample_rate
        str_root = hash_md5(str(self.root_dir)) + "_" + hash_md5(str(wav))
        wav_str = f"{str_root}_{sr}"
        if check_cache(wav_str):
            audio = load_cache(wav_str)
        else:
            audio, _ = librosa.load(wav, sr=self.sample_rate)
            cache(audio, wav_str)
        with open(self.texts[idx], "r", encoding="utf-8") as f:
            text = f.read().replace("\n", "")
        if audio.shape[0] == 0:
            print(f"Empty audio file: {wav}, padding with zeros.")
            audio = np.zeros(16000)
        audio = audio / (np.max(np.abs(audio)) + 1e-6)
        return audio, text

    def __hash__(self) -> int:
        h = hashlib.md5()
        h.update(str(self.__class__).encode())
        h.update(str(self.root_dir).encode())
        h.update(str(self.sample_params["n"]).encode())
        h.update(str(self.sample_params["seed"]).encode())
        h.update(str(True).encode())
        return int(h.hexdigest(), 16)

    def __repr__(self) -> str:
        return f"({self.root_dir.name})"


class TarDataset(Dataset):
    """
    A dataset class for a tar file containing
    with wav files and corresponding text files.
    """

    def __init__(
        self,
        root_tar: str = None,
        sample_rate: int = 22050,
        text_suffix: str = ".txt",
        path_prefix: str = None,
    ):
        super().__init__(Path(root_tar).name, sample_rate)
        if root_tar is None:
            raise ValueError("root_tar must be provided.")
        self.root_tar = root_tar
        self.root_dir = Path(root_tar).name
        self.tar = tarfile.open(root_tar)
        wavs, texts = [], []
        for member in self.tar.getmembers():
            if member.name.endswith(".wav"):
                wav_file = Path(member.name)
                wavs.append(wav_file)
                text_file = Path(member.name).with_suffix(text_suffix)
                texts.append(text_file)
        self.wavs = wavs
        self.texts = texts
        self.path_prefix = path_prefix

    def __len__(self) -> int:
        if self.indices is not None:
            return len(self.indices)
        return len(self.wavs)

    def __getitem__(self, idx: int) -> Tuple[np.ndarray, str]:
        if self.indices is not None:
            idx = self.indices[idx]
        wav, sr = self.wavs[idx], self.sample_rate
        wav_str = f"{Path(self.root_tar).name}_{wav}_{sr}"
        wav_str = wav_str.replace(".", "_")
        wav_str = wav_str.replace("/", "_")
        if check_cache(wav_str):
            audio = load_cache(wav_str)
        else:
            if self.path_prefix is not None:
                wav = self.path_prefix + str(wav)
            else:
                wav = str(wav)
            wav_file = self.tar.extractfile(wav)
            audio, _ = librosa.load(wav_file, sr=self.sample_rate)
            cache(audio, wav_str)
        if self.path_prefix is not None:
            text_f = self.path_prefix + str(self.texts[idx])
        else:
            text_f = str(self.texts[idx])
        text_file = self.tar.extractfile(text_f)
        try:
            text = text_file.read().decode("utf-8")
        except UnicodeDecodeError:
            text = ""
            print(f"Error reading text file: {text_f}")
        if audio.shape[0] == 0:
            print(f"Empty audio file: {wav}, padding with zeros.")
            audio = np.zeros(16000)
        else:
            # remove silence at beginning and end
            audio, _ = librosa.effects.trim(audio)
        return audio, text

    def __hash__(self) -> int:
        h = hashlib.md5()
        h.update(str(self.__class__).encode())
        h.update(str(self.root_tar).encode())
        h.update(str(self.sample_params["n"]).encode())
        h.update(str(self.sample_params["seed"]).encode())
        h.update(str(True).encode())
        return int(h.hexdigest(), 16)

    def __repr__(self) -> str:
        return f"({Path(self.root_tar).name})"


DEFAULT_BENCHMARKS = [
    "hubert",
    "wav2vec2",
    "wavlm",
    "wav2vec2_wer",
    "whisper_wer",
    "mpm",
    "pitch",
    "wespeaker",
    "dvector",
    "hubert_token",
    "voicefixer",
    "wada_snr",
]


class DataDistribution:
    def __init__(
        self,
        dataset: Dataset = None,
        benchmark_dict: Dict[str, "Benchmark"] = None,
        benchmarks: List[str] = DEFAULT_BENCHMARKS,
        name: str = None,
    ):
        if name is not None:
            self.name = name
        elif dataset is not None:
            self.name = dataset.name
        self.benchmarks = benchmarks
        if benchmark_dict is not None:
            self.benchmark_objects = {
                benchmark: benchmark_dict[benchmark] for benchmark in benchmarks
            }
        self.benchmark_results = {}
        if dataset is not None:
            self.dataset = dataset
            self.run()

    def run(self):
        for benchmark in self.benchmark_objects:
            print(f"Running {benchmark} on {self.dataset.root_dir}")
            bench = self.benchmark_objects[benchmark]
            dist = bench.get_distribution(self.dataset)
            if bench.dimension.name == "N_DIMENSIONAL":
                # compute mu and sigma and store as tuple
                mu = np.mean(dist, axis=0)
                sigma = np.cov(dist, rowvar=False)
                dist = (mu, sigma)
            self.benchmark_results[benchmark] = dist

    def get_distribution(self, benchmark_name: str) -> np.ndarray:
        if benchmark_name not in self.benchmark_results:
            self.run()
        return self.benchmark_results[benchmark_name]

    def to_pickle(self, path: str):
        with gzip.open(path, "wb") as f:
            pickle.dump(self.benchmark_results, f)

    @staticmethod
    def from_pickle(path: str):
        with gzip.open(path, "rb") as f:
            benchmark_results = pickle.load(f)
        obj = DataDistribution()
        obj.benchmark_results = benchmark_results
        name = Path(path).name
        if "." in name:
            name = name.split(".")[0]
        obj.name = name
        return obj



"""
This module contains functions to calculate distribution distances.
"""
import numpy as np

try:
    import jax.numpy as jnp
except ImportError:
    print("JAX is not installed. Using numpy instead, which may be slower.")
    import numpy as jnp
from scipy import linalg


def wasserstein_distance(x, y):
    """
    See: https://en.wikipedia.org/wiki/Wasserstein_metric
    """
    means = []
    np.random.seed(0)
    for _ in range(10):
        if x.shape[0] != y.shape[0]:
            # sample from the larger distribution
            if x.shape[0] > y.shape[0]:
                x = x[np.random.choice(x.shape[0], y.shape[0], replace=False)]
            else:
                y = y[np.random.choice(y.shape[0], x.shape[0], replace=False)]
        means.append(jnp.mean((jnp.sort(x) - jnp.sort(y)) ** 2) ** 0.5)
    return np.mean(means)


def frechet_distance(x, y, eps=1e-6):
    if isinstance(x, tuple):
        mu1, sigma1 = x
    else:
        mu1 = jnp.mean(x, axis=0)
        sigma1 = jnp.cov(x, rowvar=False)
    if isinstance(y, tuple):
        mu2, sigma2 = y
    else:
        mu2 = jnp.mean(y, axis=0)
        sigma2 = jnp.cov(y, rowvar=False)

    mu1 = jnp.atleast_1d(mu1)
    mu2 = jnp.atleast_1d(mu2)

    sigma1 = jnp.atleast_2d(sigma1)
    sigma2 = jnp.atleast_2d(sigma2)

    assert (
        mu1.shape == mu2.shape
    ), "Training and test mean vectors have different lengths"
    assert (
        sigma1.shape == sigma2.shape
    ), "Training and test covariances have different dimensions"

    diff = mu1 - mu2

    # Product might be almost singular
    covmean, _ = linalg.sqrtm(sigma1.dot(sigma2).astype(complex), disp=False)
    if not jnp.isfinite(covmean).all():
        msg = (
            "fid calculation produces singular product; "
            "adding %s to diagonal of cov estimates"
        ) % eps
        print(msg)
        offset = jnp.eye(sigma1.shape[0]) * eps
        covmean = linalg.sqrtm((sigma1 + offset).dot(sigma2 + offset).astype(complex))

    # Numerical error might give slight imaginary component
    if jnp.iscomplexobj(covmean):
        if not jnp.allclose(jnp.diagonal(covmean).imag, 0, atol=1e-2):
            m = jnp.max(jnp.abs(covmean.imag))
            print(f"Warning: Imaginary component {m}")
        covmean = covmean.real

    tr_covmean = jnp.trace(covmean)

    result = diff.dot(diff) + jnp.trace(sigma1) + jnp.trace(sigma2) - 2 * tr_covmean

    return result



from typing import Union

import numpy as np
import torch
from transformers import Wav2Vec2Processor, Wav2Vec2Model
from tqdm import tqdm
import librosa

from ttsds.benchmarks.benchmark import Benchmark, BenchmarkCategory, BenchmarkDimension
from ttsds.util.dataset import Dataset


class Wav2Vec2Benchmark(Benchmark):
    """
    Benchmark class for the Wav2Vec2 benchmark.
    """

    def __init__(
        self,
        wav2vec2_model: str = "facebook/wav2vec2-base",
        wav2vec2_layer: Union[int, str] = 8,
    ):
        super().__init__(
            name="Wav2Vec2",
            category=BenchmarkCategory.OVERALL,
            dimension=BenchmarkDimension.N_DIMENSIONAL,
            description="Wav2Vec2 hidden states.",
            wav2vec2_model=wav2vec2_model,
            wav2vec2_layer=wav2vec2_layer,
        )
        self.processor = Wav2Vec2Processor.from_pretrained(
            "facebook/wav2vec2-base-960h"
        )
        self.model = Wav2Vec2Model.from_pretrained(wav2vec2_model)
        self.model_layer = wav2vec2_layer

    def get_embedding(self, wav, sr) -> np.ndarray:
        """
        Get the embedding of a wav file.

        Args:
            wav (np.ndarray): The wav file.

        Returns:
            np.ndarray: The embedding of the wav file.
        """
        if sr != 16000:
            wav = librosa.resample(wav, orig_sr=sr, target_sr=16000)
            sr = 16000
        input_values = self.processor(
            wav, return_tensors="pt", sampling_rate=sr
        ).input_values
        with torch.no_grad():
            features = self.model(input_values, output_hidden_states=True).hidden_states
        if isinstance(self.model_layer, int):
            features = features[self.model_layer].detach().cpu().numpy()[0]
        else:
            layer_num = len(features)
            features_new = []
            for i in range(layer_num):
                features_new.append(features[i].detach().cpu().numpy()[0])
            features = np.stack(features_new, axis=-1)
            features = features.reshape(features.shape[0], -1)
        return features

    def _get_distribution(self, dataset: Dataset) -> np.ndarray:
        """
        Get the distribution of the Wav2Vec2 benchmark.

        Args:
            dataset (DirectoryDataset): The dataset to get the distribution from.

        Returns:
            np.ndarray: The distribution of the Wav2Vec2 benchmark.
        """
        wavs = [
            wav
            for wav, _ in tqdm(dataset, desc=f"loading wavs for {self.name} {dataset}")
        ]
        embeddings = []
        for wav in tqdm(wavs):
            embeddings.append(self.get_embedding(wav, dataset.sample_rate))
        return np.vstack(embeddings)



from typing import Union

import numpy as np
import torch
from transformers import WavLMModel, Wav2Vec2FeatureExtractor
from tqdm import tqdm
import librosa

from ttsds.benchmarks.benchmark import Benchmark, BenchmarkCategory, BenchmarkDimension
from ttsds.util.dataset import Dataset


class WavLMBenchmark(Benchmark):
    """
    Benchmark class for the WavLM benchmark.
    """

    def __init__(
        self,
        wavlm_model: str = "microsoft/wavlm-base-plus",
        wavlm_layer: Union[int, str] = 11,
    ):
        super().__init__(
            name="WavLM",
            category=BenchmarkCategory.OVERALL,
            dimension=BenchmarkDimension.N_DIMENSIONAL,
            description="WavLM hidden states.",
            wavlm_model=wavlm_model,
            wavlm_layer=wavlm_layer,
        )
        self.processor = Wav2Vec2FeatureExtractor.from_pretrained(
            "microsoft/wavlm-base-plus-sv"
        )
        self.model = WavLMModel.from_pretrained(wavlm_model)
        self.model_layer = wavlm_layer

    def get_embedding(self, wav, sr) -> np.ndarray:
        """
        Get the embedding of a wav file.

        Args:
            wav (np.ndarray): The wav file.

        Returns:
            np.ndarray: The embedding of the wav file.
        """
        if sr != 16000:
            wav = librosa.resample(wav, orig_sr=sr, target_sr=16000)
            sr = 16000
        input_values = self.processor(
            wav, return_tensors="pt", sampling_rate=sr
        ).input_values
        with torch.no_grad():
            features = self.model(input_values, output_hidden_states=True).hidden_states
        if isinstance(self.model_layer, int):
            features = features[self.model_layer].detach().cpu().numpy()[0]
        else:
            layer_num = len(features)
            features_new = []
            for i in range(layer_num):
                features_new.append(features[i].detach().cpu().numpy()[0])
            features = np.stack(features_new, axis=-1)
            features = features.reshape(features.shape[0], -1)
        return features

    def _get_distribution(self, dataset: Dataset) -> np.ndarray:
        """
        Get the distribution of the WavLM benchmark.

        Args:
            dataset (DirectoryDataset): The dataset to get the distribution from.

        Returns:
            np.ndarray: The distribution of the WavLM benchmark.
        """
        wavs = [
            wav
            for wav, _ in tqdm(dataset, desc=f"loading wavs for {self.name} {dataset}")
        ]
        embeddings = []
        for wav in tqdm(wavs):
            embeddings.append(self.get_embedding(wav, dataset.sample_rate))
        return np.vstack(embeddings)



import tempfile
import os

import wespeaker
from tqdm import tqdm
import numpy as np
import librosa
import soundfile as sf
from pyannote.audio import Model, Inference


from ttsds.benchmarks.benchmark import Benchmark, BenchmarkCategory, BenchmarkDimension
from ttsds.util.dataset import Dataset


class WeSpeakerBenchmark(Benchmark):
    """
    Benchmark class for the WeSpeaker benchmark.
    """

    def __init__(
        self,
        window_duration: float = 1.0,
        window_step: float = 0.5,
        measure_std: bool = False,
    ):
        super().__init__(
            name="WeSpeaker",
            category=BenchmarkCategory.SPEAKER,
            dimension=BenchmarkDimension.N_DIMENSIONAL,
            description="The speaker embeddings using WeSpeaker.",
            window_duration=window_duration,
            window_step=window_step,
            measure_std=measure_std,
        )
        self.model = Model.from_pretrained("pyannote/wespeaker-voxceleb-resnet34-LM")
        self.inference = Inference(
            self.model, window="sliding", duration=window_duration, step=window_step
        )
        self.measure_std = measure_std

    def _get_distribution(self, dataset: Dataset) -> np.ndarray:
        """
        Get the distribution of the WeSpeaker benchmark.

        Args:
            dataset (Dataset): The dataset to get the distribution from.

        Returns:
            np.ndarray: The distribution of the WeSpeaker benchmark.
        """
        embeddings = []
        for wav, _ in tqdm(dataset, desc=f"computing embeddings for {self.name}"):
            if dataset.sample_rate != 16000:
                wav = librosa.resample(
                    wav, orig_sr=dataset.sample_rate, target_sr=16000
                )
            with tempfile.NamedTemporaryFile(suffix=".wav") as f:
                sf.write(f.name, wav, 16000)
                embedding = self.inference(f.name)
                embedding = [x[1] for x in embedding]
            if self.measure_std:
                embedding = np.std(embedding, axis=0)
                embeddings.append(embedding)
            else:
                embeddings.extend(embedding)
        embeddings = np.vstack(embeddings)
        return embeddings



from typing import Union, List
import importlib.resources

import numpy as np
import torch
from transformers import Wav2Vec2Processor, HubertModel
from tqdm import tqdm
import librosa
from sklearn.cluster import KMeans
import requests

from ttsds.benchmarks.benchmark import Benchmark, BenchmarkCategory, BenchmarkDimension
from ttsds.util.dataset import Dataset, TarDataset
from ttsds.util.cache import cache, load_cache, check_cache, hash_md5, CACHE_DIR


class HubertTokenBenchmark(Benchmark):
    """
    Benchmark class for the Hubert benchmark.
    """

    def __init__(
        self,
        hubert_model: str = "facebook/hubert-base-ls960",
        hubert_layer: Union[int, str] = 7,
        cluster_num: int = 100,
        cluster_seed: int = 42,
        cluster_datasets: List[Dataset] = None,
    ):
        super().__init__(
            name="Hubert Token",
            category=BenchmarkCategory.PROSODY,
            dimension=BenchmarkDimension.ONE_DIMENSIONAL,
            description="Hubert hidden states.",
            hubert_model=hubert_model,
            hubert_layer=hubert_layer,
            cluster_num=cluster_num,
            cluster_seed=cluster_seed,
            cluster_datasets=cluster_datasets,
        )
        if cluster_datasets is None:
            TEST_DS_URL = "https://www.openslr.org/resources/60/test-clean.tar.gz"
            # download to cache
            TEST_DS_PATH = CACHE_DIR / "test-clean.tar.gz"
            if not TEST_DS_PATH.exists():
                print(f"downloading {TEST_DS_URL} to {TEST_DS_PATH} for HubertTokenBenchmark")
                with open(TEST_DS_PATH, "wb") as f:
                    f.write(requests.get(TEST_DS_URL).content)

            cluster_datasets = [
                TarDataset(TEST_DS_PATH, text_suffix=".normalized.txt", path_prefix="./").sample(
                    100
                )
            ]
        self.processor = Wav2Vec2Processor.from_pretrained(
            "facebook/hubert-large-ls960-ft"
        )
        self.model = HubertModel.from_pretrained(hubert_model)
        self.model_layer = hubert_layer
        self.kmeans = self.create_clusters(cluster_num, cluster_seed, cluster_datasets)

    def create_clusters(
        self, cluster_num: int, cluster_seed: int, cluster_datasets: Dataset
    ) -> KMeans:
        """
        Create clusters for the Hubert benchmark.
        """
        cache_id = self.__hash__()
        if check_cache(cache_id):
            cluster_centres = load_cache(cache_id)
            kmeans = KMeans(n_clusters=cluster_num, random_state=cluster_seed)
            dummy = np.zeros((100, 768))
            kmeans.fit(dummy)
            kmeans.cluster_centers_ = cluster_centres
            return kmeans
        wavs = []
        for ds in tqdm(cluster_datasets, desc=f"loading wavs for {self.name}"):
            wavs.extend(
                [
                    (wav, ds.sample_rate)
                    for wav, _ in tqdm(ds, desc=f"loading wavs for {self.name} {ds}")
                ]
            )
        embeddings = []
        print(len(wavs))
        for wav in tqdm(wavs):
            embeddings.append(self.get_embedding(wav[0], wav[1]))
        embeddings = np.vstack(embeddings)
        kmeans = KMeans(n_clusters=cluster_num, random_state=cluster_seed).fit(
            embeddings
        )
        cache(kmeans.cluster_centers_, cache_id)
        return kmeans

    def get_embedding(self, wav, sr) -> np.ndarray:
        """
        Get the embedding of a wav file.

        Args:
            wav (np.ndarray): The wav file.

        Returns:
            np.ndarray: The embedding of the wav file.
        """
        if sr != 16000:
            wav = librosa.resample(wav, orig_sr=sr, target_sr=16000)
            sr = 16000
        input_values = self.processor(
            wav, return_tensors="pt", sampling_rate=sr
        ).input_values
        with torch.no_grad():
            features = self.model(input_values, output_hidden_states=True).hidden_states
        features = features[self.model_layer].detach().cpu().numpy()[0]
        return features

    def _get_distribution(self, dataset: Dataset) -> np.ndarray:
        """
        Get the distribution of the Hubert benchmark.

        Args:
            dataset (DirectoryDataset): The dataset to get the distribution from.

        Returns:
            np.ndarray: The distribution of the Hubert benchmark.
        """
        wavs = [
            wav
            for wav, _ in tqdm(dataset, desc=f"loading wavs for {self.name} {dataset}")
        ]
        lengths = []
        for wav in tqdm(wavs):
            wav_emb = self.get_embedding(wav, dataset.sample_rate)
            cluster = self.kmeans.predict(wav_emb)
            # the lengths are the number of times each cluster is repeated in a row
            current_length = 1
            for i in range(1, len(cluster)):
                if cluster[i] == cluster[i - 1]:
                    current_length += 1
                else:
                    lengths.append(current_length)
                    current_length = 1
            lengths.append(current_length)
        return np.array(lengths)



import torch
import numpy as np
import librosa
from tqdm import tqdm

from ttsds.util.mpm import MaskedProsodyModel
from ttsds.util.measures import (
    PitchMeasure,
    EnergyMeasure,
    VoiceActivityMeasure,
)
from ttsds.benchmarks.benchmark import Benchmark, BenchmarkCategory, BenchmarkDimension
from ttsds.util.dataset import Dataset


class MPMBenchmark(Benchmark):
    """
    Benchmark class for the Masked Prosody Model (MPM) benchmark.
    """

    def __init__(
        self,
        mpm_model: str = "cdminix/masked_prosody_model",
        mpm_layer: int = 7,
    ):
        super().__init__(
            name="MPM",
            category=BenchmarkCategory.PROSODY,
            dimension=BenchmarkDimension.N_DIMENSIONAL,
            description="The hidden states of the Masked Prosody Model (MPM).",
            mpm_model=mpm_model,
            mpm_layer=mpm_layer,
        )
        self.model = MaskedProsodyModel.from_pretrained(mpm_model)
        self.model_layer = mpm_layer
        self.model.eval()
        self.pitch_measure = PitchMeasure()
        self.energy_measure = EnergyMeasure()
        self.voice_activity_measure = VoiceActivityMeasure()
        self.pitch_min = 50
        self.pitch_max = 300
        self.energy_min = 0
        self.energy_max = 0.2
        self.vad_min = 0
        self.vad_max = 1
        self.bins = torch.linspace(0, 1, 128)
        self.mpm_layer = mpm_layer

    def get_embedding(self, wav, sr) -> np.ndarray:
        """
        Get the embedding of a wav file.

        Args:
            wav (np.ndarray): The wav file.

        Returns:
            np.ndarray: The embedding of the wav file.
        """
        features = self.model(wav, sr)
        features = features[self.model_layer].detach().cpu().numpy()[0]
        return features

    def _get_distribution(self, dataset: Dataset) -> np.ndarray:
        """
        Get the distribution of the MPM benchmark.

        Args:
            dataset (Dataset): The dataset to get the distribution from.

        Returns:
            np.ndarray: The distribution of the MPM benchmark.
        """
        embeddings = []
        for wav, _ in tqdm(
            dataset, desc="loading masked prosody model representations"
        ):
            if dataset.sample_rate != 22050:
                wav = librosa.resample(
                    wav, orig_sr=dataset.sample_rate, target_sr=22050
                )
            pitch = self.pitch_measure(wav, np.array([1000]))["measure"]
            energy = self.energy_measure(wav, np.array([1000]))["measure"]
            vad = self.voice_activity_measure(wav, np.array([1000]))["measure"]
            pitch = torch.tensor(pitch)
            energy = torch.tensor(energy)
            vad = torch.tensor(vad)
            pitch[torch.isnan(pitch)] = -1000
            energy[torch.isnan(energy)] = -1000
            vad[torch.isnan(vad)] = -1000
            pitch = torch.clip(pitch, self.pitch_min, self.pitch_max)
            energy = torch.clip(energy, self.energy_min, self.energy_max)
            vad = torch.clip(vad, self.vad_min, self.vad_max)
            pitch = pitch / (self.pitch_max - self.pitch_min)
            energy = energy / (self.energy_max - self.energy_min)
            vad = vad / (self.vad_max - self.vad_min)
            pitch = torch.bucketize(pitch, self.bins)
            energy = torch.bucketize(energy, self.bins)
            vad = torch.bucketize(vad, torch.linspace(0, 1, 2))
            min_len = min(len(pitch), len(energy), len(vad))
            pitch = pitch[:min_len]
            energy = energy[:min_len]
            vad = vad[:min_len]
            model_input = torch.stack([pitch, energy, vad]).unsqueeze(0)
            with torch.no_grad():
                reprs = self.model(model_input, return_layer=self.mpm_layer)[
                    "representations"
                ]
            reprs = reprs.detach().cpu().numpy()[0]
            embeddings.append(reprs)
        embeddings = np.vstack(embeddings)
        return embeddings



import torch
import torch.nn as nn


class ChangeSampleRate(nn.Module):
    def __init__(self, input_rate: int, output_rate: int):
        super().__init__()
        self.output_rate = output_rate
        self.input_rate = input_rate

    def forward(self, wav: torch.tensor) -> torch.tensor:
        # Only accepts 1-channel waveform input
        wav = wav.view(wav.size(0), -1)
        new_length = wav.size(-1) * self.output_rate // self.input_rate
        indices = torch.arange(new_length) * (self.input_rate / self.output_rate)
        round_down = wav[:, indices.long()]
        round_up = wav[:, (indices.long() + 1).clamp(max=wav.size(-1) - 1)]
        output = round_down * (1.0 - indices.fmod(1.0)).unsqueeze(
            0
        ) + round_up * indices.fmod(1.0).unsqueeze(0)
        return output



from pesq import pesq
from tqdm import tqdm
import librosa
import numpy as np

from ttsds.benchmarks.benchmark import Benchmark, BenchmarkCategory, BenchmarkDimension
from ttsds.util.dataset import Dataset


class PESQBenchmark(Benchmark):
    """
    Benchmark class for the Hubert benchmark.
    """

    def __init__(
        self,
        reference_dataset: Dataset,
        sample_rate: int = 16000,
    ):
        super().__init__(
            name="PESQ",
            category=BenchmarkCategory.EXTERNAL,
            dimension=BenchmarkDimension.ONE_DIMENSIONAL,
            description="The PESQ benchmark.",
        )
        self.reference_dataset = reference_dataset
        self.sample_rate = sample_rate

    def _get_distribution(self, dataset: Dataset) -> np.ndarray:
        """
        Get the distribution of the Hubert benchmark.

        Args:
            dataset (DirectoryDataset): The dataset to get the distribution from.

        Returns:
            np.ndarray: The distribution of the Hubert benchmark.
        """
        wavs = [
            (wav, txt)
            for wav, txt, _ in tqdm(
                dataset, desc=f"loading wavs for {self.name} {dataset}"
            )
        ]
        wavs_ref = [
            (wav, txt)
            for wav, txt, _ in tqdm(
                self.reference_dataset,
                desc=f"loading wavs for {self.name} {self.reference_dataset}",
            )
        ]
        embeddings = []
        scores = []
        sr = dataset.sample_rate
        sr_ref = self.reference_dataset.sample_rate
        for (wav, txt), (wav_ref, txt_ref) in zip(wavs, wavs_ref):
            if sr != self.sample_rate:
                wav = librosa.resample(wav, orig_sr=sr, target_sr=self.sample_rate)
            if sr_ref != self.sample_rate:
                wav_ref = librosa.resample(
                    wav_ref, orig_sr=sr_ref, target_sr=self.sample_rate
                )
            if txt != txt_ref:
                raise ValueError(f"Text mismatch between {txt} and {txt_ref}")
            score = pesq(self.sample_rate, wav_ref, wav, "wb")
            scores.append(score)
        return scores



from random import sample

import torchaudio
import torch
import torch.nn as nn
import requests
import numpy as np
from tqdm import tqdm

import ttsds.benchmarks.external.utmos.lightning_module
from ttsds.benchmarks.external.utmos.change_sample_rate import ChangeSampleRate
from ttsds.benchmarks.benchmark import Benchmark, BenchmarkCategory, BenchmarkDimension
from ttsds.util.dataset import Dataset
from ttsds.util.cache import CACHE_DIR

CHECKPOINT_URL = "https://huggingface.co/spaces/sarulab-speech/UTMOS-demo/resolve/main/epoch%3D3-step%3D7459.ckpt?download=true"
CHECKPOINT_PATH = CACHE_DIR / "utmos.ckpt"

if not CHECKPOINT_PATH.exists():
    print("Downloading UTMOS checkpoint...")
    response = requests.get(CHECKPOINT_URL)
    with open(CHECKPOINT_PATH, "wb") as f:
        f.write(response.content)


class UTMOSBenchmark(Benchmark):
    """
    Benchmark class for the UTMOS benchmark.
    """

    def __init__(
        self,
    ):
        super().__init__(
            name="UTMOS",
            category=BenchmarkCategory.EXTERNAL,
            dimension=BenchmarkDimension.ONE_DIMENSIONAL,
            description="The UTMOS benchmark.",
        )
        self.model = lightning_module.BaselineLightningModule.load_from_checkpoint(
            CHECKPOINT_PATH
        )
        self.model.eval()
        self.model = self.model.to("cpu")

    def _get_distribution(self, dataset: Dataset) -> np.ndarray:
        """
        Get the distribution of the UTMOS benchmark.

        Args:
            dataset (DirectoryDataset): The dataset to get the distribution from.

        Returns:
            np.ndarray: The distribution of the UTMOS benchmark.
        """
        scores = []
        for wav, _ in tqdm(dataset, desc=f"computing scores for {self.name}"):
            # batch = torch.tensor(wav).unsqueeze(0).repeat(10, 1, 1)
            csr = ChangeSampleRate(dataset.sample_rate, 16000)
            out_wavs = csr(torch.tensor(wav).unsqueeze(0))
            batch = {
                "wav": out_wavs,
                "domains": torch.tensor([0]),
                "judge_id": torch.tensor([288]),
            }
            with torch.no_grad():
                output = self.model(batch)
            score = output.mean(dim=1).squeeze().detach().numpy() * 2 + 3
            scores.append(score)
        return scores



from abc import ABC, abstractmethod
import os

import numpy as np
import pyworld as pw
import librosa
import torch


class Measure(ABC):
    def __init__(self, name, description, is_binary=False):
        self.name = name
        self.description = description
        self.is_binary = is_binary
        if self.is_binary:
            self.name += "_binary"

    @staticmethod
    def interpolate(x):
        def nan_helper(y):
            return np.isnan(y), lambda z: z.nonzero()[0]

        nans, y = nan_helper(x)
        x[nans] = np.interp(y(nans), y(~nans), x[~nans])
        return x

    @abstractmethod
    def compute(self, audio, durations, silence_mask=None):
        pass

    def __call__(self, audio, durations, silence_mask=None, include_prior=False):
        measure = self.compute(audio, durations, silence_mask)
        if include_prior:
            return {
                "measure": Measure.interpolate(measure),
                "prior": {
                    "mean": measure[~np.isnan(measure)].mean(),
                    "std": measure[~np.isnan(measure)].std(),
                },
            }
        else:
            return {"measure": Measure.interpolate(measure)}

    def __eq__(self, other):
        return self.name == other.name


class PitchMeasure(Measure):
    def __init__(
        self,
        name="pitch",
        description="Pitch measure",
        sampling_rate=22050,
        hop_length=256,
        pitch_quality=1,
    ):
        global pw
        super().__init__(name, description)
        self.sampling_rate = sampling_rate
        self.hop_length = hop_length
        self.dio_speed = int(np.round(1 / pitch_quality))

    def compute(self, audio, durations, silence_mask=None):
        f0, t = pw.dio(
            audio.astype(np.float64),
            self.sampling_rate,
            frame_period=self.hop_length / self.sampling_rate * 1000,
            speed=self.dio_speed,
        )
        pitch = pw.stonemask(
            audio.astype(np.float64), f0, t, self.sampling_rate
        ).astype(np.float32)
        if sum(durations) < len(pitch):
            pitch = pitch[: sum(durations)]
        if silence_mask is not None:
            pitch[silence_mask] = np.nan
        pitch[pitch < 1e-6] = np.nan
        if np.isnan(pitch).all():
            pitch[:] = 1e-6
        return pitch


class VoiceActivityMeasure(Measure):
    def __init__(
        self,
        name="voice_activity",
        description="Voice activity measure",
        sampling_rate=22050,
        win_length=1024,
        hop_length=256,
        pitch_quality=1,
    ):
        global pw
        super().__init__(name, description, is_binary=True)
        self.sampling_rate = sampling_rate
        self.hop_length = hop_length
        self.dio_speed = int(np.round(1 / pitch_quality))

    def compute(self, audio, durations, silence_mask=None):
        f0, t = pw.dio(
            audio.astype(np.float64),
            self.sampling_rate,
            frame_period=self.hop_length / self.sampling_rate * 1000,
            speed=self.dio_speed,
        )
        pitch = pw.stonemask(
            audio.astype(np.float64), f0, t, self.sampling_rate
        ).astype(np.float32)
        if sum(durations) < len(pitch):
            pitch = pitch[: sum(durations)]
        if silence_mask is not None:
            pitch[silence_mask] = np.nan
        pitch[pitch < 1e-6] = np.nan
        pitch[~np.isnan(pitch)] = 1
        pitch[np.isnan(pitch)] = 0
        return pitch


class EnergyMeasure(Measure):
    def __init__(
        self,
        name="energy",
        description="Energy measure",
        win_length=1024,
        hop_length=256,
    ):
        super().__init__(name, description)
        self.win_length = win_length
        self.hop_length = hop_length

    def compute(self, audio, durations, silence_mask=None):
        energy_overall = np.sum(audio ** 2) / len(audio)
        energy = librosa.feature.rms(
            y=audio,
            frame_length=self.win_length,
            hop_length=self.hop_length,
            center=True,
        ).astype(np.float32)
        energy = energy.reshape(-1)
        if sum(durations) < len(energy):
            energy = energy[: sum(durations)]
        return energy



import re
import tempfile

import whisper
from tqdm import tqdm
from jiwer import wer
import torch
import numpy as np
import librosa
import soundfile as sf

from ttsds.benchmarks.benchmark import Benchmark, BenchmarkCategory, BenchmarkDimension
from ttsds.util.dataset import Dataset


class WhisperWERBenchmark(Benchmark):
    """
    Benchmark class for the Whisper Word Error Rate (WER) benchmark.
    """

    def __init__(
        self,
        whisper_model: str = "small.en",
    ):
        super().__init__(
            name="Whisper WER",
            category=BenchmarkCategory.INTELLIGIBILITY,
            dimension=BenchmarkDimension.ONE_DIMENSIONAL,
            description="The Word Error Rate (WER) of Whisper.",
            whisper_model=whisper_model,
        )
        self.model = whisper.load_model(whisper_model)

    def _get_distribution(self, dataset: Dataset) -> np.ndarray:
        """
        Compute the Word Error Rate (WER) distribution of the Whisper model.

        Args:
            dataset (Dataset): The dataset to compute the WER on.

        Returns:
            float: The Word Error Rate (WER) distribution of the Whisper model.
        """
        wers = []
        for wav, gt_transcript in tqdm(dataset, desc=f"computing WER for {self.name}"):
            if dataset.sample_rate != 16000:
                wav = librosa.resample(
                    wav, orig_sr=dataset.sample_rate, target_sr=16000
                )
            with tempfile.NamedTemporaryFile(suffix=".wav") as f:
                sf.write(f.name, wav, 16000)
                pred_transcript = self.model.transcribe(f.name)["text"]
            pred_transcript = re.sub(r"[^\w\s]", "", pred_transcript)
            gt_transcript = re.sub(r"[^\w\s]", "", gt_transcript)
            pred_transcript = re.sub(r"\s+", " ", pred_transcript)
            gt_transcript = re.sub(r"\s+", " ", gt_transcript)
            pred_transcript = pred_transcript.strip().lower()
            gt_transcript = gt_transcript.strip().lower()
            if len(gt_transcript) == 0:
                wers.append(1.0)
            else:
                wers.append(wer(gt_transcript, pred_transcript))
        return np.array(wers)



import tempfile

from pesq import pesq
from voicefixer import VoiceFixer
from simple_hifigan import Synthesiser
import numpy as np
import soundfile as sf
import librosa
from tqdm import tqdm

from ttsds.benchmarks.benchmark import Benchmark, BenchmarkCategory, BenchmarkDimension
from ttsds.util.dataset import Dataset


class VoiceFixerBenchmark(Benchmark):
    """
    Benchmark class for the VoiceFixer benchmark.
    """

    def __init__(
        self,
    ):
        super().__init__(
            name="VoiceFixer",
            category=BenchmarkCategory.ENVIRONMENT,
            dimension=BenchmarkDimension.ONE_DIMENSIONAL,
            description="The phone counts of VoiceFixer.",
        )
        self.model = VoiceFixer()
        self.synthesiser = Synthesiser()

    def _get_distribution(self, dataset: Dataset) -> np.ndarray:
        """
        Compute the Word Error Rate (WER) distribution of the VoiceFixer model.

        Args:
            dataset (Dataset): The dataset to compute the WER on.

        Returns:
            float: The Word Error Rate (WER) distribution of the VoiceFixer model.
        """
        mel_diffs = []
        for wav, _ in tqdm(dataset, desc=f"computing noise for {self.name}"):
            if dataset.sample_rate != 16000:
                wav = librosa.resample(
                    wav, orig_sr=dataset.sample_rate, target_sr=16000
                )
            with tempfile.NamedTemporaryFile(suffix=".wav") as f:
                # take random 2 seconds
                if len(wav) > 32000:
                    start = np.random.randint(0, len(wav) - 32000)
                    wav = wav[start : start + 32000]
                sf.write(f.name, wav, 16000)
                with tempfile.NamedTemporaryFile(suffix=".wav") as f_out:
                    self.model.restore(f.name, f_out.name)
                    wav_out, _ = librosa.load(f_out.name, sr=16000)
            wav = wav / (np.max(np.abs(wav)) + 1e-5)
            wav_out = wav_out / np.max(np.abs(wav_out))
            mel = self.synthesiser.wav_to_mel(wav, 16000)[0].T
            mel_out = self.synthesiser.wav_to_mel(wav_out, 16000)[0].T
            if mel_out.shape[0] > mel.shape[0]:
                mel_out = mel_out[: mel.shape[0]]
            elif mel_out.shape[0] < mel.shape[0]:
                mel = mel[: mel_out.shape[0]]
            mel_diff = mel_out
            # check if there is any nan
            if np.isnan(mel_diff).any():
                print("nan found, skip")
                continue
            # convert back to wav
            mel_diff = self.synthesiser(mel_diff.T)[0]
            mel_diff = mel_diff / np.max(np.abs(mel_diff) + 1e-5)
            mel_diff = librosa.resample(mel_diff, orig_sr=22050, target_sr=16000)
            # calculate the difference
            try:
                mel_diff = pesq(16000, wav, mel_diff, "wb")
                mel_diffs.append(mel_diff)
            except:
                mel_diffs.append(0)
        return np.array(mel_diffs)
